from tqdm import tqdm_notebook
from itertools import product as iterprod
from pytearcat.tensor.misc import new_ten,reload_all
from pytearcat.tensor.core import config
from pytearcat.tensor.tensor import tensor_series, Tensor
from pytearcat.tensor.core.core import core_calc, display_IP, Math_IP, Latex_IP

if core_calc == 'sp':

    from pytearcat.tensor.core.core import diff, Rational, factor

elif core_calc == 'gp':

    from pytearcat.tensor.core.core import diff, divide, simplify,series

def D(element,i):

    '''
    Hasta ahora esta funcion deriva cualquier cosa del tipo "MutableDenseNDimArray" con respecto a una sola coordenada, ingresada en forma de numero.
    '''

    #print("tipo de funcion derivada",type(diff),diff)

    coord_index = config.coords

    # FIRST CASE: VECTOR DERIVATIVE WITH RESPECT TO ONE COORDINATE  d/dx^m Tensor[_i,^j,^k,_l,...etc]
    # WHERE x^m CORRESPOND TO i

    exec(reload_all('config'),globals(),locals()) # Import all the variables from config

    if type(i) == int:

        #coordinate = sympify(coord_index[i])

        str12 = "valor = diff(element,%s)"%(str(coord_index[i]))

        

        #str12 = "valor = diff(element,%s)"%str(coord_index[i])

        #str12 = str12.replace('XXXX',str(coordinate))

        #str12 = str12.replace('XXXX',str(coord_index[i]))

        exec(str12,locals(),globals())

        return valor

    else:

        print("ERROR IN THE DERIVATIVE D OF A TENSOR")

class ChristoffelClass(Tensor):

    def __init__(self):

        super().__init__('Christoffel',3)

    def __call__(self,str_index):

        lista = str_index.split(',')

        updn = [symbol[0] for symbol in lista]

        if updn[1] == '^' or updn[2] == '^':

            raise SyntaxError("Christoffel can be only first or second kind.")

        else:

            return super().__call__(str_index)

    def display(self, index=None, aslist = None):

        if index is None:

            super().display("_,_,_",aslist)
            super().display("^,_,_",aslist)
            
        elif index == "^,_,_" or index == "_,_,_":

            super().display(index,aslist)

        else:

            raise SyntaxError("Christoffel can be only first or second kind.")




def calculate_christoffel(First_kind=True,Second_kind=True):
    
    global Christoffel
    
    g = config.g
    var = list(config.coords.values())
    dim = config.dim
    ord_status = config.ord_status
    coords = config.coords

    

    if config.christ is None:

        Christoffel = config.create_ten("Christoffel",ChristoffelClass())

        config.christ = Christoffel

    else:

        Christoffel = config.christ

    # True if the first kind (second kind) symbol is already calculated. False if not.

    calc_fk = Christoffel.indices[0] 

    calc_sk = Christoffel.indices[4]

    

    description = r'Christoffel '


    if First_kind == True and Second_kind==True:

        if calc_fk == True:

            description += r'$\Gamma^{\alpha}_{\beta \gamma}$'

        elif calc_sk == True:

            description += r'$\Gamma_{\alpha \beta \gamma}$'

        else:

            description += r'$\Gamma_{\alpha \beta \gamma}$ and $\Gamma^{\alpha}_{\beta \gamma}$'

    elif First_kind == True:

        description += r'$\Gamma_{\alpha \beta \gamma}$'

    elif Second_kind == True:

        description += r'$\Gamma^{\alpha}_{\beta \gamma}$'

    coord_index = coords
    exec(reload_all('config'),globals(),locals()) # Import all the variables from config


    # Check if the program really needs to calculate a Christoffel Symbol

    if First_kind == False and Second_kind == False:

        print('You need to ask for at least one kind of Christoffel Symbol.')

        return Christoffel
    

    elif calc_fk == True and First_kind == True and Second_kind == False:

        #print('First Kind Christoffel already calculated.')

        display_IP(Latex_IP(r"First Kind Christoffel Symbol $\Gamma_{\alpha \beta \gamma}$ already calculated"))

        return Christoffel

    elif calc_sk == True and Second_kind == True and First_kind == False:

        #print('Second Kind Christoffel already calculated.')

        display_IP(Latex_IP(r"Second Kind Christoffel Symbol $\Gamma^{\alpha}_{\beta \gamma}$ already calculated"))

        return Christoffel

    elif calc_fk == True and calc_sk == True:

        #print('First and Second kind Christoffel already calculated.')

        display_IP(Latex_IP(r"First and Second Kind Christoffel Symbol $\Gamma_{\alpha \beta \gamma}$ and $\Gamma^{\alpha}_{\beta \gamma}$ already calculated"))

        return Christoffel

    if core_calc == 'gp':

        display_IP(Latex_IP(description))

    elif core_calc == 'sp':

        display_IP(Latex_IP(description))

    for p in tqdm_notebook(iterprod(range(dim),repeat=3),total=dim**3):

        if p[2] >= p[1]:

            countm = p[0]
            counti = p[1]
            countj = p[2]
            
            if First_kind==True and not(calc_fk):  #First Kind ('_,_,_')

                if core_calc == 'sp':
                
                    FirstTemp = Rational(1,2)*(diff(g.tensor[0][countm][countj],var[counti])+diff(g.tensor[0][countm][counti],var[countj])-diff(g.tensor[0][counti][countj],var[countm]))
                
                elif core_calc == 'gp':
                    
                    FirstTemp = divide(1,2)[0]*(diff(g.tensor[0][countm][countj],var[counti])+diff(g.tensor[0][countm][counti],var[countj])-diff(g.tensor[0][counti][countj],var[countm]))
                
                if ord_status == True:

                    Christoffel.tensor[0][countm][counti][countj] = tensor_series(FirstTemp)

                else:

                    if core_calc == 'sp':

                        Christoffel.tensor[0][countm][counti][countj] = factor(FirstTemp)

                    elif core_calc == 'gp':

                        Christoffel.tensor[0][countm][counti][countj] = simplify(FirstTemp)

                # Simetria
                Christoffel.tensor[0][countm][countj][counti] = Christoffel.tensor[0][countm][counti][countj]

                Christoffel.indices[0] = True

            if Second_kind==True and not(calc_sk):  #Second Kind ('^,_,_')
                
                SecondTemp = 0
                    
                for countk in range(dim):
                    
                    if core_calc == 'sp':

                        SecondTemp += Rational(1,2)*g.tensor[3][countk][countm]*(diff(g.tensor[0][counti][countk],var[countj]) + diff(g.tensor[0][countj][countk],var[counti]) - diff(g.tensor[0][counti][countj],var[countk]) )
                
                    elif core_calc == 'gp':

                        SecondTemp += divide(1,2)[0]*g.tensor[3][countk][countm]*(diff(g.tensor[0][counti][countk],var[countj]) + diff(g.tensor[0][countj][countk],var[counti]) - diff(g.tensor[0][counti][countj],var[countk]) )

                if ord_status == True:

                    Christoffel.tensor[4][countm][counti][countj] = tensor_series(SecondTemp)

                else:
                    
                    if core_calc == 'sp':

                        Christoffel.tensor[4][countm][counti][countj] = factor(SecondTemp)

                    elif core_calc == 'gp':

                        Christoffel.tensor[4][countm][counti][countj] = simplify(SecondTemp)

                # Simetria
                Christoffel.tensor[4][countm][countj][counti] = Christoffel.tensor[4][countm][counti][countj]

                Christoffel.indices[4] = True

    Christoffel.space()

    return Christoffel


from itertools import product as iterprod
from pytearcat.tensor.core import config
from pytearcat.tensor.core.core import *
from pytearcat.tensor.misc import new_var, reload_all
from pytearcat.tensor.tensor import Tensor, tensor_series
from pytearcat.tensor.core.config import *

if core_calc == 'gp':

    import io
    from contextlib import redirect_stdout
    from pytearcat.tensor.core.display import gp_pretty_latex


def create_metric(ds2 = ''):

    global g

    if config.g_status == True:

        answer = input('The metric has already been defined. Do you want to overwrite it? (yes/no)')

        if answer == 'yes' or answer == 'y' or answer == 'Y' or answer == 'YES':

            print('The metric has been redefined. Standard tensors have been deleted.')

            config.g_status = False

            config.g = None

            config.christ = None

            config.ricci = None

            config.ricciS = None

            config.riemann = None

            config.G = None

            config.__all__.remove('g')

    if core_calc == 'sp':

        coords = ','.join(list(map(str,config.coords.values())))

    elif core_calc == 'gp':

        temp = ''

        for i in list(config.coords.values()):
            
            lol = "%s"%str(i)

            temp = temp[:] + lol[:] + ','
            
        coords = temp[:-1]

    g_matrix,line_element = metric(config.fun, coords, ds2)

    display_ds(g_matrix)
    
    if core_calc == 'gp':

        f = io.StringIO()
        with redirect_stdout(f):
            print(latex(matrix(g_matrix)))
        out = f.getvalue()

        out = out.replace(r"\\",r"\\\\").replace("\\text{","").replace("\"}\"","").replace('\"','').replace('\\}','}').replace('\\{','{')#.replace('\\\\','\\')

        display_IP(Math_IP(gp_pretty_latex(r"%s"%out)))

    elif core_calc == 'sp':

        display(Matrix(g_matrix))

    return config.g

def display_ds(gmatrix):

    dim = config.dim
    coords = config.coords.copy()
    greek = config.greek_dict

    for i in coords:

        if core_calc == 'sp':

            if str(coords[i]) in list(greek.keys()):

                coords[i] = greek[str(coords[i])]

        elif core_calc == 'gp':

            gpcoord = tolatex(coords[i]).replace("\\","")

            if gpcoord in list(greek.keys()):

                coords[i] = greek[gpcoord]

            del(gpcoord)

    ds = r'ds^2 = '

    for i,j in iterprod(range(dim),repeat=2):

        value = gmatrix[i,j]

        if value != 0:            

            if i == j:

                if core_calc == 'gp':

                    ds += r"\left(%s\right)*d%s^2+"%(tolatex(value),coords[i])

                elif core_calc == 'sp':

                    if value.is_Add:

                        ds += r"\left(%s\right)*d%s^2+"%(latex(value),coords[i])

                    else:
                    
                        ds += r"%s*d%s^2+"%(latex(value),coords[i])

            elif j > i:

                if core_calc == 'gp':

                    ds += r"\left(%s\right)*d%s*d%s+"%(tolatex(2*value),coords[i],coords[j])

                elif core_calc == 'sp':

                    if value.is_Add:

                        ds += r"\left(%s\right)*d%s*d%s+"%(latex(value*2),coords[i],coords[j])

                    else:

                        ds += r"%s*d%s*d%s+"%(latex(value*2),coords[i],coords[j])

    ds = ds[:-1]

    ds = ds.replace(r'+-',r'-')
    ds = ds.replace(r'++',r'+')
    ds = ds.replace(r'**',r'^')
    ds = ds.replace(r'*',r' \cdot ')

    if core_calc == 'gp':

        display_IP(Math_IP(gp_pretty_latex(ds)))
        
    elif core_calc == 'sp':

        display_IP(Math_IP(ds))
        

def metric(functions, coords, ds2):

    '''
    It creates the metric matrix. Receives te functions defined (as sympy functions),
    the coordinates separated by comma and the line element ds^2.
    '''   
    # if coords == '':

    #     # If the user does not enter the coordinates, it asks for them.
        
    #     variables_string = coordinates() 
    #     config.dim = len(variables_string.split(','))
    #     config.coords = coord_index(variables_string)
        
    # else:
        
    #     variables_string = coords
    #     config.dim = len(variables_string.split(','))
    #     config.coords = coord_index(variables_string)
    
    variables_string = coords

    if ds2 == '':
        
        ds2 = input('Enter the metric in the form: ds2=-dt*dt+a**2*(dx**2+dy**2+dz**2): \n' )
    
        config.ds = ds2 ####--!!! Guarda el string ds en el archivo config.
    
    print('\nDimension = %d'%config.dim)
    
    print('Coordinates = %s'%variables_string)
    
    g_matrix, line_element = create_metric_matrix(config.dim, variables_string, ds2)  

    print('Metric defined successfully: \n')    
    
    return g_matrix, line_element

def create_metric_matrix(dim, variables_string, ds_input):
    
    # CREAMOS UNA VARIABLE POR CADA DIMENSION Y VERIFICAMOS QUE CALCEN, 
    # QUEDAN ASIGNADAS CON LAS LETRAS QUE SE INGRESARON COMO INPUT 
    
    for i in range(0,dim):

        COORDENADA = variables_string.split(',')

        if len(COORDENADA) != dim:
            raise ValueError('The number of coordinates and dimension must be equal.')

        # HERE WE CREATE THE DIFFERENTIAL COORDINATES WITH THE INPUT NAMES: e.g. dt,dx,dy,dz

        if core_calc == 'gp':

            string = "d%s = giac('d%s')" % (COORDENADA[i],COORDENADA[i])
        
        elif core_calc == 'sp':

            string = "d%s = Symbol('d%s')" % (COORDENADA[i],COORDENADA[i])

        exec(string,locals(),globals())
    
    #HERE WE ASSIGN A INDEX NUMBER TO EVERY COORDINATE CREATED BY DE USER 
    # IT IS ENUMERATED FROM 0 TO  DIM-1
    # THE INDEX COORDINATE CAN BE FOUND BY coord_index['x']
    
    exec(reload_all('config'),globals(),locals()) # Import all the variables from config

    # We execute the string ds3 that will define the line element for the metric
    
    exec(ds_input,locals(),globals())
    
    g_matrix=zeros(dim, dim)  #g_matrix=(row i, column j)

    c = config.coords
    
    for i,j in iterprod(range(dim),repeat=2):
        
        if core_calc == 'sp':

            factor='d%s*d%s'%(c[i],c[j])

            element = expand(ds2)
            
            coef = eval(r"element.coeff(%s)"%factor,locals(),globals())

            coef = nsimplify(coef)

        elif core_calc == 'gp':

            factor0='d%s'%c[i]
            factor1='d%s'%c[j]


        if i!=j:

            if core_calc == 'gp':

                coef = eval(r"simplify(coeff(coeff(ds2,%s,1),%s,1))"%(factor0,factor1),locals(),globals())


            g_matrix[i,j] =coef/2
            
        else:

            if core_calc == 'gp':

                coef = eval(r"simplify(coeff(ds2,%s,2))"%factor0,locals(),globals())

            g_matrix[i,j] =coef 
    
    g_matrix_inv = g_matrix.inv()

    if config.ord_status == True:

        for i,j in iterprod(range(dim),repeat=2):

            g_matrix[i,j] = tensor_series(g_matrix[i,j])
            g_matrix_inv[i,j] = tensor_series(g_matrix_inv[i,j])

    g = config.create_ten('g',Tensor('g',2))

    # SOLO EL TENSOR g uv TENDRA DOS OPCIONES DD Y UU, NO HAY DU Y UD PORQUE ESOS SIEMPRE SERAN IDENTIDAD, LUEGO
    if core_calc == 'gp':

        identidad = idn(dim)

    elif core_calc == 'sp':

        identidad = eye(dim)

    for p in iterprod(range(dim),repeat=2):

        i = p[0]
        j = p[1]

        if j >= i:

            g.tensor[0][i][j] = g_matrix[i,j]
            g.tensor[3][i][j] = g_matrix_inv[i,j]
            g.tensor[1][i][j] = identidad[i,j]
            g.tensor[2][i][j] = identidad[i,j]

            # Simetria
            g.tensor[0][j][i] = g.tensor[0][i][j]
            g.tensor[3][j][i] = g.tensor[3][i][j]
            g.tensor[1][j][i] = identidad[i,j]
            g.tensor[2][j][i] = identidad[i,j]
    
    config.g_status = True
    
    config.g = g

    config.g.space()
    
    return g_matrix, ds2



def ask_dim():
    
    config.dim = int(input("Enter a the dimension:\n"))

    return config.dim


def coord_index(coordinates):

    '''
    It defines a dictionary { 0 : 'x' , 1 : 'y' ...}
    
    '''

    dim = config.dim
    
    coord = coordinates.split(',')

    a = [None]*dim  # ARRAY OF INDEX
    b = [None]*dim  # ARRAY OF COORDINATES

    for i in range(0,dim):

        a[i] = i
        b[i] = new_var(coord[i])

    return dict(zip(a,b))

def def_coords(coords):

    '''
    It defines the coordinates of the manifold. 

    coords = 't,x,y,z'
    
    '''

    if not isinstance(coords,str):

        raise TypeError("The argument must be a string.")

    print('Remember that the time coordinate must be the first coordinate.')

    coordinates = []

    variables_string = coords

    config.dim = len(variables_string.split(','))

    config.coords = coord_index(variables_string)

    config.coords_sp = {i-1 : config.coords[i] for i in config.coords if i != 0}

    coordinates.append(new_var(*coords.split(',')))

    if len(coordinates) == 1:

        return coordinates[0]

    else:

        return coordinates



from tqdm import tqdm_notebook
from itertools import product as iterprod 

from pytearcat.tensor.misc import new_ten,reload_all
from pytearcat.tensor.core.core import factor, display_IP, Latex_IP,core_calc,simplify
from pytearcat.tensor.core import config
from pytearcat.tensor.tensor import construct, tensor_series, Tensor

from .christoffel import calculate_christoffel, D

def calculate_riemann(default = True, All = False):

    # Riemann (_,_,_,_)

    global Riemann

    dim = config.dim

    if config.riemann is None:

        #Riemann = new_ten('Riemann',4)

        Riemann = config.create_ten("Riemann",Tensor('Riemann',4))

        config.riemann = Riemann

    else:

        Riemann = config.riemann

    if config.christ is None:

        Christ_local = calculate_christoffel()

    else:

        Christ_local = config.christ
    
    Ch1 = Christ_local.tensor[0][:][:][:] # First kind Christoffel
    Ch2 = Christ_local.tensor[4][:][:][:] # Second Kind Christoffel

    gmet = config.g.tensor

    if Riemann.indices[0] == False and default == False:

        Riemann_list = construct('False',dim,4)
        
        Riem0 = Riemann.tensor[0]

        #wolfram alpha = counta ; beta = countb ; gamma = countc ; delta = countd

        display_IP(Latex_IP(r'Riemann Tensor $R_{\alpha \beta \gamma \delta}$'))

        for p in tqdm_notebook(iterprod(range(dim),repeat=4),total=dim**4):

            counta = p[0]
            countb = p[1]
            countc = p[2]
            countd = p[3]

            if Riemann_list[counta][countb][countc][countd] == False:

                Value = 0
                Right = 0

                for countk in range(dim):

                    Value += (gmet[0][counta][countk]*(D(Ch2[countk][countb][countd],countc) - D(Ch2[countk][countb][countc],countd)))#.simplify()

                    Right += (Ch2[countk][countb][countd]*Ch1[counta][countk][countc] - Ch2[countk][countb][countc]*Ch1[counta][countk][countd])#.simplify()

                if core_calc == 'gp':

                    Rie_val = simplify(Value + Right)                   

                elif core_calc == 'sp':

                    Rie_val = factor((Value + Right))

                if config.ord_status == True:

                    Riem0[counta][countb][countc][countd] = tensor_series(Rie_val)

                else:

                    Riem0[counta][countb][countc][countd] = Rie_val

                Riemann_list[counta][countb][countc][countd] = True 

                # Skew Symmetry

                if Riemann_list[counta][countb][countd][countc] == False:

                    #print('Skew1',[counta,countb,countc,countd],[counta,countb,countd,countc])

                    Riem0[counta][countb][countd][countc] = -Riem0[counta][countb][countc][countd]

                    Riemann_list[counta][countb][countd][countc] = True

                if Riemann_list[countb][counta][countc][countd] == False:
                    
                    #print('Skew2',[counta,countb,countc,countd],[countb,counta,countc,countd])

                    Riem0[countb][counta][countc][countd] = -Riem0[counta][countb][countc][countd]

                    Riemann_list[countb][counta][countc][countd] = True

                # Interchange Symmetry

                if Riemann_list[countc][countd][counta][countb] == False:
                    
                    #print('Interchange',[counta,countb,countc,countd],[countc,countd,counta,countb])

                    Riem0[countc][countd][counta][countb] = Riem0[counta][countb][countc][countd]

                    Riemann_list[countc][countd][counta][countb] = True

                # Bianchi Identity (First)

                if Riemann_list[counta][countc][countd][countb] == False and Riemann_list[counta][countd][countb][countc] == True:

                    #print('Bianchi1',[counta,countb,countc,countd],[counta,countc,countd,countb])

                    Riem0[counta][countc][countd][countb] = -Riem0[counta][countb][countc][countd] - Riem0[counta][countd][countb][countc]

                    Riemann_list[counta][countc][countd][countb] = True

                if Riemann_list[counta][countd][countb][countc] == False and Riemann_list[counta][countc][countd][countb] == True:
                    
                    #print('Bianchi2',[counta,countb,countc,countd],[counta,countd,countb,countc])

                    Riem0[counta][countd][countb][countc] = -Riem0[counta][countb][countc][countd] - Riem0[counta][countc][countd][countb]

                    Riemann_list[counta][countd][countb][countc] = True

        Riemann.indices[0] = True

    #######################################

    if Riemann.indices[8] == False:

        Riemann_list = construct('False',dim,4)

        Riem8 = Riemann.tensor[8]
        #wolfram alpha = counta ; beta = countb ; gamma = countc ; delta = countd

        display_IP(Latex_IP(r'Riemann Tensor $R^{\alpha}_{\beta \gamma \delta}$'))

        for p in tqdm_notebook(iterprod(range(dim),repeat=4),total=dim**4):

            counta = p[0]
            countb = p[1]
            countc = p[2]
            countd = p[3]

            #print(Riemann_list)

            if Riemann_list[counta][countb][countc][countd] == False:

                Rie_val = 0

                Value = D(Ch2[counta][countb][countd],countc) - D(Ch2[counta][countb][countc],countd)

                Right = 0

                for count_mu in range(dim):

                    Right += Ch2[count_mu][countb][countd]*Ch2[counta][count_mu][countc] - Ch2[count_mu][countb][countc]*Ch2[counta][count_mu][countd]
                
                Rie_val = Value + Right


                if config.ord_status == True:

                    Riem8[counta][countb][countc][countd] = tensor_series(Rie_val)

                else:

                    if core_calc == 'gp':

                        Rie_val = simplify(Rie_val)                   

                    elif core_calc == 'sp':

                        Rie_val = factor(Rie_val)

                    Riem8[counta][countb][countc][countd] = Rie_val

                Riemann_list[counta][countb][countc][countd] = True 
                
                # Skew Symmetry
            
                if Riemann_list[counta][countb][countd][countc] == False:

                    #print('Skew1')

                    Riem8[counta][countb][countd][countc] = -Riem8[counta][countb][countc][countd]

                    Riemann_list[counta][countb][countd][countc] = True

        Riemann.indices[8] = True

    else:

        display_IP(Latex_IP(r"Riemann Tensor $R_{\alpha \beta \gamma \delta}$ already calculated"))

    if All == True:

        Riemann.complete('^,_,_,_')

    Riemann.space()

    return Riemann


from tqdm import tqdm_notebook
from itertools import product as iterprod
from pytearcat.tensor.misc import new_var, new_fun
from pytearcat.tensor.core import config
from pytearcat.tensor.tensor import tensor_series
from pytearcat.tensor.core.core import core_calc
from .christoffel import calculate_christoffel

if core_calc == 'sp':

    from pytearcat.tensor.core.core import diff, Rational, sympify, factor

elif core_calc == 'gp':

    from pytearcat.tensor.core.core import diff, divide as Rational

def geodesic():

    global geo

    dim = config.dim

    coords = config.geocoords

    if config.geo is None:

        geo = []

        config.geo = geo

    else:

        print("Geodesic already calculated.")

        geo = config.geo

        return geo, coords

    if config.christ is None:

        Christ_local = calculate_christoffel()

    else:

        Christ_local = config.christ
    
    Ch2 = Christ_local.tensor[4][:][:][:] # Second Kind Christoffel

    s = new_var('s') # proteger la variable s
    
    confcoords = list(config.coords.values())

    for i, elem in enumerate(confcoords):
        
        coords[elem] = new_fun('x__%d'%i,'s')

    geocoords = list(coords.values())

    for i,j,k in tqdm_notebook(iterprod(range(dim),repeat=3),total=dim**3):

        if (j+k) == 0:

            result = diff(geocoords[i],s,2)
        
        result += Ch2[i][j][k] * diff(geocoords[j],s) * diff(geocoords[k],s)

        if (j+k) == 2*(dim-1):

            if config.ord_status == True:

                result = tensor_series(result)

            else:

                result = sympify(factor(result))

            geo.append(result)

    return geo, coords






from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["jupyter>=1.0.0", "numpy==1.19.0", "sympy>=1.7.1", "tqdm>=4.55.0", "jedi==0.17.2"]

ext_requirements = {"giacpy" : ["giacpy>=0.7.0"]}

setup(
    name="pytearcat",
    version="0.0.4",
    author="Joaquin Sureda",
    author_email="jmsureda@uc.cl",
    description="A package to perform tensor algebra calculations.",
    long_description=readme,
    long_description_content_type="text/markdown",
    licence="LICENCE",
    url="https://github.com/pytearcat/pytearcat",
    packages=find_packages(),
    install_requires=requirements,
    extras_require=ext_requirements,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
)


from .metric import def_coords, create_metric
from .christoffel import calculate_christoffel as christoffel
from .riemann import calculate_riemann as riemann
from .ricci import calculate_ricci as ricci
from .ricci import calculate_ricci_scalar as ricci_scalar
from .einstein import calculate_einstein as einstein
from .geodesic import geodesic




class gError(Exception):

    pass


class TensorSyntaxError(Exception):

    def __init__(self, message,):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)



from re import findall,search
from pytearcat.tensor.core import config
from .core import get_name, latex, display_IP, Math_IP, core_calc

if core_calc == "gp":

    from .core import tolatex, expand
    from .gpwrap import gpcore

    def gp_pretty_order(element):
        
        ord_n = config.ord_n
        ord_var = get_name(config.ord_var)
        
        if ord_var in config.greek_dict:

            ord_var = r"\%s"%ord_var

        structure = r"\%s\^\{(\d+)\} \\operatorname\{\\mathrm\{order\\_size\} \}\\left\(\%s\\right\)"%(ord_var,ord_var)

        exponent = findall(structure,element)

        if len(exponent) > 0:

            structure = r"%s^{%s} \operatorname{\mathrm{order\_size} }\left(%s\right)"%(ord_var,exponent[0],ord_var)

            final = element.replace(r"%s"%structure,r"\mathrm{O}\left(%s^{%s}\right)"%(ord_var,exponent[0]))
            
        elif len(exponent) == 0:

            structure = r"%s \operatorname{\mathrm{order\_size} }\left(%s\right)"%(ord_var,ord_var)

            final = element.replace(r"%s"%structure,r"\mathrm{O}\left(%s\right)"%ord_var)

        return final

    def gp_pretty_latex(element):
        
        '''
        It takes a giacpy latex expression and returns a latex string that is similar to the sympy notation.
        
            - It rewrites the derivatives
            - It rewrites the expansion order
        
        '''
        
        names = {}
        variables = []

        result = element[:]
        
        greek_dict =config.greek_dict

        for i in config.fun:

            x = r'%s'%str(latex(i))[1:].split("\left")[0].replace('\\','').replace('{','\\{').replace('}','\\}').replace('mathrm','\\\\mathrm')

            names[x] = str(i).split('(')[1][:-1]

        for i in names:

            j = findall(r'(?<=\W)%s\^\{\\left\((.*?)\\right\)\}'%i,element)
            
            for k in j:

                ind = k.split(',')

                ini,fin = search(r'%s\^\{\\left\(%s\\right\)\}'%(i,k),element).span()
                
                while element[ini] != ' ' and element[ini] != ',' and element[ini] != '\\' and element[ini] != '-' and element[ini] != '+':

                    ini -= 1

                string = ''

                l_dict = {}

                for l in ind: 

                    l_dict[l] = ind.count(l)  
                    
                for l in l_dict:  # L = '1'

                    y = names[i].split(',')[int(l)-1]
                        
                    if y in greek_dict.keys():
                            
                        y = r'\\%s'%y
                    
                    if l_dict[l] == 1:
                        
                        string += r'\frac{\partial }{\partial %s}'%y

                    else:

                        string += r'\frac{\partial^%d }{\partial %s^%d}'%(l_dict[l],y,l_dict[l])

                string+= i

                result = result.replace(element[ini:fin],string)    

        for func in names:

            if func in greek_dict.keys():

                result = result.replace(func,greek_dict[func])
                
        if config.ord_status == True:

            result = gp_pretty_order(result)

        return result.replace('\\\\','\\').replace("\"","").replace("\\{","{").replace("\\}","}")


        def tolatex(element):

            f = io.StringIO()
            with redirect_stdout(f):
                print(latex(element))
            element = f.getvalue()

            string =  element[1:-2]

            return gp_pretty_latex(string)

    def display(a , b=None):    
        '''
        a is an element from giacpy
        b is a string e.g. "T^{a}_{b}^{c}"

        '''

        if not isinstance(a,gpcore):

            raise TypeError("The argument must be a single element, not a list or array.")

        if config.ord_status == True:

            a = expand(a)

        string = tolatex(a)

        string = gp_pretty_latex(string)

        if b is not None:

            string = "%s = %s"%(b,string)

        display_IP(Math_IP(string))

if core_calc == 'sp':

    from pytearcat.tensor.core import core
    
    def display(a,b = None):

        '''
        a is an element from sympy
        b is an string e.g. "T^{a}_{b}^{c}"
        
        '''

        if not isinstance(a,core.Expr):

            raise TypeError("The argument must be a single element, not a list or array.")

    
        string = str(latex(a))

        if b is not None:

            string = "%s = %s"%(b,string)
    
        display_IP(Math_IP(string))
